"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
var index_1 = require("typedoc/dist/lib/models/reflections/index");
var components_1 = require("typedoc/dist/lib/converter/components");
var converter_1 = require("typedoc/dist/lib/converter/converter");
var CommentPlugin_1 = require("typedoc/dist/lib/converter/plugins/CommentPlugin");
var models_1 = require("typedoc/dist/lib/models");
/**
 * A handler that deals with inherited reflections.
 */
var NoInheritPlugin = /** @class */ (function (_super) {
    __extends(NoInheritPlugin, _super);
    function NoInheritPlugin() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    /**
     * Create a new CommentPlugin instance.
     */
    NoInheritPlugin.prototype.initialize = function () {
        this.listenTo(this.owner, converter_1.Converter.EVENT_BEGIN, this.onBegin);
        this.listenTo(this.owner, converter_1.Converter.EVENT_CREATE_DECLARATION, this.onDeclaration, -100); // after CommentPlugin
        this.listenTo(this.owner, converter_1.Converter.EVENT_RESOLVE_BEGIN, this.onBeginResolve);
    };
    /**
     * Triggered when the converter begins converting a project.
     *
     * @param context  The context object describing the current state the converter is in.
     */
    NoInheritPlugin.prototype.onBegin = function (context) {
        this.noInherit = [];
        this.inheritedReflections = [];
    };
    /**
     * Triggered when the converter has created a declaration or signature reflection.
     *
     * Builds the list of classes/interfaces that don't inherit docs and
     * the list of reflections that are inherited that could end up being removed.
     *
     * @param context  The context object describing the current state the converter is in.
     * @param reflection  The reflection that is currently processed.
     * @param node  The node that is currently processed if available.
     */
    NoInheritPlugin.prototype.onDeclaration = function (context, reflection, node) {
        if (reflection instanceof index_1.DeclarationReflection) {
            // class or interface that won't inherit docs
            if (reflection.kindOf(index_1.ReflectionKind.ClassOrInterface) &&
                reflection.comment && reflection.comment.hasTag('noinheritdoc')) {
                this.noInherit.push(reflection);
                CommentPlugin_1.CommentPlugin.removeTags(reflection.comment, 'noinheritdoc');
            }
            // class or interface member inherited from a super
            if (reflection.inheritedFrom && reflection.parent && reflection.parent.kindOf(index_1.ReflectionKind.ClassOrInterface) &&
                (!reflection.overwrites || (reflection.overwrites && reflection.overwrites !== reflection.inheritedFrom))) {
                this.inheritedReflections.push(reflection);
            }
        }
    };
    /**
     * Triggered when the converter begins resolving a project.
     *
     * Goes over the list of inherited reflections and removes any that are down the hierarchy
     * from a class that doesn't inherit docs.
     *
     * @param context The context object describing the current state the converter is in.
     */
    NoInheritPlugin.prototype.onBeginResolve = function (context) {
        var _this = this;
        if (this.noInherit) {
            var project_1 = context.project;
            var removals_1 = [];
            this.inheritedReflections.forEach(function (reflection) {
                var resolvedInherit = _this.resolveType(context, reflection, reflection.inheritedFrom);
                // Look up the inheritance chain for a super that doesn't inherit this reflection
                if (_this.isNoInheritUpHierarchy(context, reflection, resolvedInherit)) {
                    removals_1.push(reflection);
                }
            });
            removals_1.forEach(function (removal) {
                CommentPlugin_1.CommentPlugin.removeReflection(project_1, removal);
            });
        }
    };
    /**
     * Takes some ReferenceType and resolve it to a reflection.
     * This is needed because we are operating prior to the TypePlugin resolving types.
     * @param context  The context object describing the current state the converter is in.
     * @param reflection  The reflection context.
     * @param type  The type to find relative to the reflection.
     */
    NoInheritPlugin.prototype.resolveType = function (context, reflection, type) {
        var project = context.project;
        if (type instanceof models_1.ReferenceType) {
            if (type.symbolID === models_1.ReferenceType.SYMBOL_ID_RESOLVE_BY_NAME) {
                return reflection.findReflectionByName(type.name);
            }
            else if (!type.reflection && type.symbolID !== models_1.ReferenceType.SYMBOL_ID_RESOLVED) {
                return project.reflections[project.symbolMapping[type.symbolID]];
            }
            else {
                return type.reflection;
            }
        }
    };
    /**
     * Checks whether some reflection's inheritance chain is broken by a class or interface that doesn't inherit docs.
     * @param context  The context object describing the current state the converter is in.
     * @param current  The current reflection being evaluated for non-inheritance.
     * @param end  The end of the inheritance chain.
     */
    NoInheritPlugin.prototype.isNoInheritUpHierarchy = function (context, current, end) {
        if (current === end)
            return false;
        // As we move up the chain, check if the reflection parent is in the noInherit list
        var parent = current.parent;
        for (var i = 0; i < this.noInherit.length; i++) {
            var no = this.noInherit[i];
            if (no.id === parent.id && no.name === parent.name) {
                return true;
            }
        }
        if (parent.extendedTypes) {
            for (var i = 0; i < parent.extendedTypes.length; i++) {
                var extended = this.resolveType(context, parent, parent.extendedTypes[i]);
                var upLevel = extended.findReflectionByName(current.name);
                if (this.isNoInheritUpHierarchy(context, upLevel, end)) {
                    return true;
                }
            }
        }
        return false;
    };
    NoInheritPlugin = __decorate([
        components_1.Component({ name: 'no-inherit' })
    ], NoInheritPlugin);
    return NoInheritPlugin;
}(components_1.ConverterComponent));
exports.NoInheritPlugin = NoInheritPlugin;
